#include "parser.hpp"
#include "split.hpp"

#include <iostream>
#include <fstream>
#include <string>
#include <cctype>
#include <algorithm>
#include <ciso646>
#include <cstdint>
#include <array>
#include <stdexcept>
#include <type_traits>
#include <climits>
#include <boost/variant/apply_visitor.hpp>
#include <sha1/sha1.hpp>
#include <sha1/BufferSource.hpp>

namespace {
std::string load_file (const std::string& path) {
	std::ifstream ifs(path);
	ifs >> std::noskipws;

	return std::string(
		std::istreambuf_iterator<char>(ifs),
		std::istreambuf_iterator<char>()
	);
}

bool is_printable(std::string_view s) {
    return std::find_if(s.cbegin(), s.cend(),
		 [](unsigned char c){ return not std::isprint(c); }
	) == s.cend();
}

struct PrintVisitor {
	PrintVisitor(int depth=0, bool skip_first=true) :
		depth(depth), first_line(skip_first), skip_first(skip_first) {}

	void print_tabs(int extra=0) {
		if (!first_line) {
			for (int z = 0; z < depth+extra; ++z) std::cout << '\t';
		}
		else {
			first_line = false;
		}
	}

	void reset() {
		first_line = skip_first;
	}

	void operator() (long long value) {
		print_tabs();
		std::cout << value << '\n';
	}

	void operator() (const duck::TorrentStringType& value) {
		constexpr std::size_t max_size = 41;

		print_tabs();

		std::string_view str{value};
		std::string_view ellip;
		if (is_printable(value)) {
			if (value.size() > max_size + 3) {
				str = str.substr(0, max_size);
				ellip = "...";
			}
		}
		else {
			str = "<binary_data>";
		}

		std::cout << "\"" << str << ellip << "\"[" << value.size() << "]\n";
	}

	void operator() (const std::vector<duck::TorrentValue>& value) {
		print_tabs();
		std::cout << "[\n";

		PrintVisitor new_visitor(depth+1, false);
		for (const auto& item : value) {
			boost::apply_visitor(new_visitor, item);
		}

		print_tabs();
		std::cout << "]\n";
	}

	void operator() (const std::map<duck::TorrentStringType, duck::TorrentValue>& value) {
		print_tabs();
		std::cout << "{\n";

		PrintVisitor new_visitor(depth+1);
		for (const auto& item : value) {
			print_tabs(1);
			std::cout << item.first << " => ";

			new_visitor.reset();
			boost::apply_visitor(new_visitor, item.second);
		}
		print_tabs();
		std::cout << "}\n";
	}

	int depth;
	bool first_line;
	bool skip_first;
};

template <typename T>
struct FindTVisitor : boost::static_visitor<T> {
	FindTVisitor (std::string_view search_path) {
		auto split = dincore::split(search_path, '/', true, true);
		search.reserve(split.size());
		for (const auto& itm : split) {
			search.emplace_back(itm.data(), itm.size());
		}
	}

	void reset() {
		search_error = false;
		search_match_count = 0;
	}

	bool is_path_found() const {
		return search.size() == search_match_count;
	}

	T operator() (long long value) {
		if constexpr (std::is_same<long long, T>::value)
			return value;
		search_error = true;
		return {};
	}

	T operator() (const duck::TorrentStringType& value) {
		if constexpr (std::is_same<duck::TorrentStringType, T>::value) {
			if (is_path_found())
				return value;
		}
		search_error = true;
		return {};
	}

	T operator() (const std::vector<duck::TorrentValue>&) {
		search_error = true;
		return {};
	}

	T operator() (const std::map<duck::TorrentStringType, duck::TorrentValue>& value) {
		if (is_path_found()) {
			search_error = true;
			return {};
		}

		auto it_found = value.find(search[search_match_count]);
		if (value.end() == it_found)
			return {};

		++search_match_count;
		return boost::apply_visitor(*this, it_found->second);
	}

	std::vector<std::string> search;
	std::size_t search_match_count{0};
	bool search_error{false};
};

std::vector<std::array<std::uint32_t, 5>> collect_hashes (std::string_view hashes) {
	typedef std::array<std::uint32_t, 5> array_type;

	if (hashes.empty())
		return {};

	constexpr std::size_t hash_size = 20;
	constexpr std::size_t uint_count = std::tuple_size<array_type>();
	static_assert(sizeof(array_type::value_type) * uint_count == hash_size);

	if (hashes.size() % hash_size != 0)
		throw std::runtime_error("Bad pieces array size " + std::to_string(hashes.size()));

	std::vector<array_type> retval(hashes.size() / hash_size);
	const char* src = hashes.data();
	for (std::size_t z = 0; z < retval.size(); ++z, src+=hash_size) {
		//std::copy_n(src, hash_size, reinterpret_cast<char*>(retval[z].data()));
		for (std::size_t n = 0; n < uint_count; ++n) {
			char* const out_uint = reinterpret_cast<char*>(retval[z].data() + n);
			for (std::size_t u = 0; u < sizeof(std::uint32_t); ++u) {
				out_uint[u] = src[n * sizeof(std::uint32_t) + sizeof(std::uint32_t) - u - 1];
			}
		}
	}
	return retval;
}

template <typename T>
T find_item(std::string_view path, const std::vector<duck::TorrentValue>& values) {
	FindTVisitor<T> visitor(path);
	for (const auto& value : values) {
		visitor.reset();
		T found = boost::apply_visitor(visitor, value);
		if (not visitor.search_error)
			return found;
	}
	return {};
}

SHA1::MessageDigest to_hash160_digest (const std::array<std::uint32_t, 5>& arr) {
	SHA1::MessageDigest hash_found;
	std::copy_n(arr.begin(), 5, hash_found.hash);
	return hash_found;
}
} //unnamed namespace

int main(int argc, const char* argv[]) {
	using duck::parse_torrent;
	typedef duck::TorrentStringType torrent_string;

	if (argc != 2) {
		std::cerr << "Wrong number of parameters. Usage:\n"
			<< argv[0] << " <path_to_torrent>\n";
		return 2;
	}

	std::string full_torrent = load_file(argv[1]);
	std::cout << "Loaded file into string of size " << full_torrent.size() << '\n';
	std::vector<duck::TorrentValue> values = parse_torrent(full_torrent);

	for (const auto& value : values) {
		PrintVisitor visitor;
		boost::apply_visitor(visitor, value);
	}

	auto source_hashes = find_item<torrent_string>("/info/pieces", values);
	std::cout << "Got source_hashes with size " << source_hashes.size() << '\n';
	auto hashes = collect_hashes(source_hashes);
	std::cout << "Got " << hashes.size() << " hashes\n";

	std::string file_name = std::string{find_item<torrent_string>("/info/name", values)};
	std::cout << "Found file name \"" << file_name << "\"\n";

	{
		const auto piece_length = find_item<long long int>("/info/piece length", values);
		std::ifstream istream(file_name, std::ios::in|std::ios::binary);

		if (istream.is_open()) {
			std::vector<std::uint8_t> buff(piece_length);
			std::size_t hash_index = 0;
			std::size_t match_count = 0;

			while (istream.read(reinterpret_cast<char*>(buff.data()), piece_length).gcount() > 0) {
				SHA1::BufferSource buff_source(buff.data(), istream.gcount() * CHAR_BIT);
				SHA1::MessageDigest calculated_hash = SHA1::computeFromSource(buff_source);
				SHA1::MessageDigest stored_hash = to_hash160_digest(hashes[hash_index]);
				std::cout << stored_hash.toHexString() << "  " <<
					calculated_hash.toHexString() << "  " << file_name << '\n';
				++hash_index;
				if (stored_hash == calculated_hash)
					++match_count;
			}

			std::cout << "Hash verified " << match_count << '/' << hashes.size() << '\n';
		}
		else {
			std::cout << "Error: opening \"" << file_name << "\" failed\n";
		}
	}
}
